<?php
namespace kingfisher\functional;

function curry($func) {
  $rf = get_reflection($func);

  $num_args = $rf->getNumberOfRequiredParameters();
  $num_args_supplied = func_num_args() - 1;

  if($num_args_supplied > $num_args) {
    $allowed =
      $num_args <= 1? 'none': 'maximum of ' . ($num_args - 1);
    throw new \Exception(
      sprintf(
        'Too many arguments: %s provided, %s allowed here',
        $num_args_supplied,
        $allowed
      )
    );
  }

  return
    $num_args_supplied === 0?
      new Curried($func, array($num_args)):
      new Curried(
        $func,
        array(
          array_slice(func_get_args(), 1),
          $num_args - $num_args_supplied
        )
      );
}

function spatter($func) {
  $rc = new \ReflectionClass(ns('Curried'));
  return
    $rc->newInstance(
      $func,
      array_slice(func_get_args(), 1),
      FALSE
    );
}

function safe_spatter($func) {
  $rc = new \ReflectionClass(ns('Curried'));
  return $rc->newInstance($func, array_slice(func_get_args(), 1));
}

class Curried {
  private $func;
  private $num_args;
  private $args = array();
  private $null_permitted = array();
  private $must_be_array = array();
  private $ref_indices = array();
  private $arg_class = array();

  function __construct($func, $arg_spec, $check_num_args = TRUE) {
    $this->func = $func;

    $rf = get_reflection($func);

    if($rf instanceof \ReflectionMethod) {
      foreach (array('abstract', 'private', 'protected') as $adjective) {
        $m = "is$adjective";
        if($rf->$m()) {
          throw new \Exception("$adjective methods are not allowed");
        }
      }
    }

    $params = $rf->getParameters();

    for($i = 0; $i < count($params); $i++) {
      $this->null_permitted[$i] =
        $rf->isInternal() ||
          (
            $params[$i]->isDefaultValueAvailable() &&
            is_null($params[$i]->getDefaultValue())
          );
        /*<
          why do we call isInternal?  the reason is given in
          the PHP docs:

            Due to implementation details, it is not possible to get the
            default value of built-in functions or methods of built-in
            classes.
        */

      $this->must_be_array[$i] = $params[$i]->isArray();

      if($params[$i]->isPassedByReference())
        array_push($this->ref_indices, $i);

      $rc = $params[$i]->getClass();
      if(is_object($rc))
        $this->arg_class[$i] = $rc->name;
    }

    if($check_num_args) {
      $num_args_implied = 0;
      foreach ($arg_spec as $x) {
        $num_args_implied += is_array($x)? count($x): intval($x);
      }

      $np = $rf->getNumberOfParameters();
      $np_req = $rf->getNumberOfRequiredParameters();

      $quantity = $num_args_implied < $np_req? 'too few':
        ($num_args_implied > $np? 'too many': NULL);

      if(is_string($quantity)) {
        throw new \Exception(
          "function has been specified with $quantity arguments: " .
            "number specified: $num_args_implied; number required: " .
            ($np === $np_req? $np: "between $np_req and $np")
        );
      }
    }

    $pos = 0;

    foreach ($arg_spec as $x) {
      if(is_array($x)) {
        /*
          we could check that the array is non-empty; but i don't
          think it would be right to enforce this.
        */
        foreach ($x as $arg) {
          $this->set_arg($pos, $arg);
          $pos++;
        }
      } else
        $pos += intval($x);
    }

    $this->num_args = $pos;
  }

  function __invoke() {
    $copy = clone $this;

    $index = 0;
    foreach (func_get_args() as $arg) {
      while(array_key_exists($index, $copy->args))
        $index++;
      $copy->set_arg($index, $arg);
    }

    switch(cmp(count($copy->args), $copy->num_args)) {
      case -1:
        return $copy;
      break;

      case 0:
        ksort($copy->args, SORT_NUMERIC);

        if(0) {
          print_r(
            array('debug: about to call', $copy->func, $copy->args)
          );
        }

        $tmp = array();
        foreach ($this->ref_indices as $index)
          $tmp[$index] = &$copy->args[$index];
        #unset($tmp);

        $result = call_user_func_array($copy->func, $copy->args);

        if(0) {
          var_dump(
            array('debug: about to return', $result)
          );
        }

        return $result;
      break;

      case 1:
        throw new \Exception(
          sprintf(
            'Too many arguments accumulated: %s expected, %s provided',
            $copy->num_args,
            count($copy->args)
          )
        );

        /*
          design decision:
            we could try to be "helpful", and use the extra
            arguments as optional arguments, if the underlying
            function has a sufficient number of
            optional arguments.  however, this would
            be mistaken.  consider a function `f' which has one
            required and one optional argument.  then, with
            the behaviour just described, this would work:
              $x = curry('f');
              $result = $x('some-arg1', 'some-arg2');
            whereas this would probably cause a crash:
              $x = curry('f');
              $x = $x('some-arg1');
              $x = $x('some-arg2');
        */
      break;
    }
  }

  function set_arg($index, $arg) {
    $null_permitted = !empty($this->null_permitted[$index]);

    if(!(is_null($arg) && $null_permitted)) {
      $msg_start = "argument $index (counting from 0) must be";

      if($null_permitted) {
        $msg_start .= ' NULL or';
      }

      if(!empty($this->must_be_array[$index]) &&
        !is_array($arg))
      {
        throw new \Exception(
          "$msg_start an array"
        );
      }

      $class = val_or_default($this->arg_class, $index, NULL);

      if(!empty($class) &&
        (!is_object($arg) || !($arg instanceof $class)))
      {
        throw new \Exception(
          "$msg_start an instance of $class"
        );
      }
    }

    $this->args[$index] = $arg;
  }
}

function get_reflection($f) {
  if(is_array($f)) {
    return new \ReflectionMethod($f[0], $f[1]);
  } elseif(is_string($f) &&
    preg_match('@^([^:]+)::(.+)@s', $f, $matches))
  {
    return new \ReflectionMethod($matches[1], $matches[2]);
  } else
    return new \ReflectionFunction($f);
}

function ns($s) {
  return __NAMESPACE__ . "\\$s";
}

function val_or_default(array $array, $key, $default) {
  return array_key_exists($key, $array)? $array[$key]: $default;
}

function cmp($a, $b) {
  return $a < $b? -1: ($a == $b? 0: 1);
}
